// 001 u_c_calcul_is
// 23 jan 2020

// -- (C) John COLIBRI 2020
// -- documentation: http://www.jcolibri.com

(*$r+*)

unit u_c_calcul_is;
  interface
    uses Classes, u_c_basic_object;

    type c_calcul_is=
             class(c_basic_object)
               m_title: String;
               m_tranche_is_count: integer;
               m_seuil_is_array, m_taux_is_array: array of Double;

               m_c_detail_result_list: tStrings;

               m_c_key_list, m_c_value_list: tStringList;

               m_capital, m_pourcent_capital: Double;
               m_resultat_moins_salaires: Double;
               m_micro_foncier_brut: Double;
               m_taux_interets_compte_associe, m_interets_compte_associe: Double;

               m_dividendes: Double;
               m_revenu_sas_IR: Double;

               m_IS, m_taux_is_moyen: Double;

               m_benefice_avant_is, m_benefice_net, m_reserve: Double;

               m_benefice_net_precedent: Double;

               Constructor create_calcul_is(p_name: String);

               procedure load_is_parameters_from_file(p_path, p_file_name: String);
               function f_amount(p_key: String): Double;
               function f_rate(p_key: String): Double;
               procedure display_is_parameters;

               procedure calcule_is_sur_revenu_imposable;

               Destructor Destroy; Override;
             end; // c_calcul_is

  implementation
    uses SysUtils, Math
        , u_c_display
        , u_c_stringlist

        , u_utils
        ;

    const k_max= 10e15;

    // -- c_calcul_is

    Constructor c_calcul_is.create_calcul_is(p_name: String);
      begin
        Inherited create_basic_object(p_name);
        m_c_key_list:= tStringList.Create;
        m_c_value_list:= tStringList.Create;
      end; // create_calcul_is

    procedure c_calcul_is.load_is_parameters_from_file(p_path, p_file_name: String);

      procedure _add_key_value(p_key, p_value: String);
        begin
          display('add '+ p_key+ ' '+ p_value);
          m_c_key_list.Add(p_key);
          m_c_value_list.Add(p_value);
        end; // _add_key_value

      var l_seuil, l_taux: String;
          l_key, l_value: String;

      begin // load_is_parameters_from_file
        // -- a first bogus (0, 0) slice
        m_tranche_is_count:= 1;
        SetLength(m_seuil_is_array, m_tranche_is_count);
        SetLength(m_taux_is_array, m_tranche_is_count);

        with c_stringlist.create_stringlist('', p_path, p_file_name) do
        begin
          stop_if_empty;

          while not f_end_of_file do
          begin
            display(m_the_line);
            if not f_is_blank_or_comment
              then begin
                  if Pos('tranche_is_', m_the_line)> 0
                    then begin
                        f_next_value;
                        l_seuil:= f_next_value; l_taux:= f_next_value;
                        Inc(m_tranche_is_count);
                        SetLength(m_seuil_is_array, m_tranche_is_count);
                        SetLength(m_taux_is_array, m_tranche_is_count);
                        if l_seuil= '+'
                          then m_seuil_is_array[m_tranche_is_count- 1]:= k_max
                          else m_seuil_is_array[m_tranche_is_count- 1]:= StrToFloat(l_seuil);
                        m_taux_is_array[m_tranche_is_count- 1]:= StrToFloat(l_taux);
                      end
                    else begin
                        l_key:= f_next_value; l_value:= f_next_value;
                        _add_key_value(l_key, l_value);
                      end;
                end
              else
                if m_title= ''
                  then m_title:= m_trimmed_line
                  else _add_key_value('','');
            read_line;
          end; // while not f_end_of_file

          Free;
        end; // with c_stringlist
      end; // load_is_parameters_from_file

    function c_calcul_is.f_amount(p_key: String): Double;
      var l_index_of: integer;
      begin
        l_index_of:= m_c_key_list.IndexOf(p_key);
        if l_index_of< 0
          then begin
              display_bug_stop('no_key '+ p_key);
              Result:= 0;
            end
          else Result:= StrToFloat(m_c_value_list[l_index_of]);
      end; // f_amount

    function c_calcul_is.f_rate(p_key: String): Double;
      var l_index_of: integer;
      begin
        l_index_of:= m_c_key_list.IndexOf(p_key);
        if l_index_of< 0
          then begin
              display_bug_stop('no_key '+ p_key);
              Result:= 0;
            end
          else Result:= StrToFloat(m_c_value_list[l_index_of]);
      end; // f_rate

    procedure c_calcul_is.display_is_parameters;
      var l_tranche_is: Integer;
          l_is_parameter_index: Integer;
          l_key: String;
      begin
        display(m_title);

        for l_tranche_is:= 0 to m_tranche_is_count- 1 do
          display(Format('tranche_is_%d %10.2f %5.2f',
            [l_tranche_is, m_seuil_is_array[l_tranche_is], m_taux_is_array[l_tranche_is] ]));
        display_line;

        with m_c_key_list do
          for l_is_parameter_index:= 0 to Count- 1 do
          begin
            l_key:= Strings[l_is_parameter_index];
            if l_key= ''
              then display_line
              else display(Format('%-25s %10.2f', [l_key, f_amount(l_key)]));
          end;
      end; // display_is_parameters

    procedure c_calcul_is.calcule_is_sur_revenu_imposable;
      var l_c_save_display_strings: tStrings;

          l_taux_deductible, l_interets_associes_deductibles, l_interets_associes_non_deductibles: Double;
                    
          l_tranche_is: Integer;
          l_base, l_taux_tranche: Double;
          l_is_tranche: Double;
          l_display: String;
      begin
        if m_benefice_avant_is> 0
          then begin
              display_line;

              l_c_save_display_strings:= g_c_display.m_c_strings;
              initialize_display(m_c_detail_result_list);
              // clear_display;

              // if m_interets_compte_associe, f_rate('taux_interet_moyen_pret')
              m_interets_compte_associe:= m_taux_interets_compte_associe* m_capital;

              l_taux_deductible:= Min(m_taux_interets_compte_associe, f_rate('taux_interet_moyen_pret')/ 100);
              l_interets_associes_deductibles:= l_taux_deductible* m_capital;
              l_interets_associes_non_deductibles:= Max(0, m_interets_compte_associe- l_interets_associes_deductibles);

              if l_interets_associes_non_deductibles> 0
                then begin
                    display(f_display('interets_associ_deduct'  , l_interets_associes_deductibles));
                    display(f_display('interets_non_deduc'       , l_interets_associes_non_deductibles));
                    display_line;
                  end;

              display(f_display('resu_moins_salaires'     , m_benefice_avant_is));
              display_line;

              m_benefice_avant_is:= m_benefice_avant_is- m_micro_foncier_brut- l_interets_associes_deductibles;

              display(f_display(' - micro_foncier'           , m_micro_foncier_brut));
              // display(f_display(' - interets_associ         %10.2f', m_interets_compte_associe));
              display(f_display(' - interets_associ_dduct' , l_interets_associes_deductibles));
              // display(f_display(' - dividendes               %10.2f', m_dividendes));
              display(f_display(' = benef_soumis_a_is'       , m_benefice_avant_is));

              // display(f_display('imposable_a_IS'             , m_benefice_avant_is));
              display_line;

              display('--- >');
              m_IS:= 0;
              for l_tranche_is:= 1 to m_tranche_is_count- 1 do
              begin
(*
                display_line;
                display(f_display('seuil %d %10.2f', l_tranche_is, m_seuil_is_array[l_tranche_is] ));
*)
                l_base:= m_benefice_avant_is- m_seuil_is_array[l_tranche_is- 1];
                if l_base> 0
                  then begin
                      l_base:= Min(m_benefice_avant_is, m_seuil_is_array[l_tranche_is]);
                      // display(f_display('base_MIN %10.2f seuil-1 %10.2f', l_base, m_seuil_is_array[l_tranche_is- 1]));
                      l_base:= l_base- m_seuil_is_array[l_tranche_is- 1];
                      // display(f_display('base_MIN %10.2f ', l_base));

                      l_taux_tranche:= m_taux_is_array[l_tranche_is] ;
//                      display(f_display('rate %10f', l_taux_tranche ));
                      l_is_tranche:= l_base* l_taux_tranche / 100;
//                      display(f_display('is %10f', l_is));
                      display(Format('is_1 %10.2f %10.2f  %5.2f   %10.2f',
                          [m_seuil_is_array[l_tranche_is], l_base, l_taux_tranche, l_is_tranche]));

                      m_IS:= m_IS+ l_is_tranche;
                    end;
              end; // for l_tranche_is

              display_line;
              display(Format('total_is      %s %10.2f  %5.2f %%  ',
                  [f_spaces(21), m_IS, m_IS/ m_benefice_avant_is]));
              m_benefice_net:= m_benefice_avant_is- m_IS;
              display('--- <');
              display_line;

              l_display:= f_display('benefices_net',   m_benefice_net);
              if m_benefice_net_precedent<> 0
                then l_display:= l_display+ Format('%s [%10.2f]', [f_spaces(2), m_benefice_net- m_benefice_net_precedent]);
              display(l_display);

              initialize_display(l_c_save_display_strings);

              display_line;
              display(f_display('is',                  m_IS));
              display_line;
            end
          else begin
              l_c_save_display_strings:= g_c_display.m_c_strings;
              initialize_display(m_c_detail_result_list);
              clear_display;

              display(f_display('*** revenu_avant_is< 0 %10.2', m_benefice_avant_is));

              initialize_display(l_c_save_display_strings);
            end;
      end; // calcule_is_sur_revenu_imposable

    Destructor c_calcul_is.Destroy;
      begin
        m_c_key_list.Free;
        m_c_value_list.Free;

        Inherited;
      end; // Destroy

    begin // u_c_calcul_is
    end. // u_c_calcul_is


