// 001 u_c_calcul_ir
// 24 jan 2020

// -- (C) John COLIBRI 2020
// -- documentation: http://www.jcolibri.com

(*$r+*)

unit u_c_calcul_ir;
  interface
    uses Classes, u_c_basic_object
        , u_strings, u_stringlist, u_c_stringlist
        , u_c_calcul_is
        ;

    type c_calcul_ir=
             class(c_stringlist)
               m_c_calcul_is_ref: c_calcul_is;
               m_c_detail_result_list: tStrings;

               m_title: String;
               m_month_count: Integer;

               m_tranche_count: Integer;
               m_seuil_array, m_taux_array: array of double;

               m_c_key_list, m_c_value_list: tStringList;

               m_retraites: Double;
               m_revenu_imposable: Double; // entered from form or from simulateur revenu
               m_nombre_parts: Double;
               m_csg_deductible_ir: Double;

               m_IR, m_ir_precedent: Double;
               m_total_csg_crds, m_total_csg_crds_precedent: Double;

               // -- check our feuille d'imposition
               m_revenu_au_taux_forfaitaire: Double;
               m_delta_net: Double;

               Constructor create_calcul_ir(p_name, p_path, p_file_name: String);

               function f_rate(p_key: String): Double;
               function f_amount(p_key: String): Double;

               procedure load_parameters;
               procedure display_parameters;

               procedure compute_ir;

               Destructor Destroy; Override;
             end; // c_calcul_ir

  implementation
    uses SysUtils, Math
        , u_c_display, u_utils
        ;

    // -- c_calcul_ir

    const k_max= 10e15;

    function f_is_tranche(p_key: String): Boolean;
      begin
        Result:= Copy(p_key, 1, 8)= 'tranche_';
      end; // f_is_tranche

    Constructor c_calcul_ir.create_calcul_ir(p_name, p_path, p_file_name: String);
      begin
        Inherited create_stringlist(p_name, p_path, p_file_name);

        m_c_key_list:= tStringList.Create;
        m_c_value_list:= tStringList.Create;

        m_nombre_parts:= 1;
      end; // create_calcul_ir

    procedure c_calcul_ir.load_parameters;

      procedure _add_key_value(p_key, p_value: String);
        begin
          display('add '+ p_key+ ' '+ p_value+ '<');
(*
          if p_key= '+'
            then p_value:= -1;
*)
          m_c_key_list.Add(p_key);
          m_c_value_list.Add(p_value);
        end; // _add_key_value

      var l_key, l_seuil, l_taux, l_value: String;

      begin // load_parameters
        stop_if_empty;
        display(m_c_stringlist.Text);

        // -- a first bogus tranche "0 0"
        m_tranche_count:= 1;
        SetLength(m_seuil_array, m_tranche_count);
        SetLength(m_taux_array, m_tranche_count);

        while not f_end_of_file do
        begin
          // display(m_the_line);
          if not f_is_blank_or_comment
            then begin
                // NO _add_key_value(f_next_value, f_next_value);
                l_key:= f_next_value;

                if f_is_tranche(l_key)
                  then begin
                      l_seuil:= f_next_value; l_taux:= f_next_value;
                      Inc(m_tranche_count);
                      SetLength(m_seuil_array, m_tranche_count);
                      SetLength(m_taux_array, m_tranche_count);
                      if l_seuil= '+'
                        then m_seuil_array[m_tranche_count- 1]:= k_max
                        else m_seuil_array[m_tranche_count- 1]:= StrToFloat(l_seuil);
                      m_taux_array[m_tranche_count- 1]:= StrToFloat(l_taux);

                      l_value:= l_seuil+ ' '+ l_taux;
                    end
                  else begin
                      // -- taux forfaitaire
                      l_value:= f_next_value;
                    end  ;

                _add_key_value(l_key, l_value);
              end
            else
              if m_title= ''
                then m_title:= m_trimmed_line
                else _add_key_value('','');

          read_line;
        end; // while not f_end_of_file
      end; // load_parameters

    function c_calcul_ir.f_rate(p_key: String): Double;
      var l_index_of: integer;
      begin
        l_index_of:= m_c_key_list.IndexOf(p_key);
        if l_index_of< 0
          then begin
              display_bug_stop('no_key '+ p_key);
              Result:= 0;
            end
          else Result:= StrToFloat(m_c_value_list[l_index_of]);
      end; // f_rate

    function c_calcul_ir.f_amount(p_key: String): Double;
      var l_index_of: integer;
      begin
        l_index_of:= m_c_key_list.IndexOf(p_key);
        if l_index_of< 0
          then begin
              display_bug_stop('no_key '+ p_key);
              Result:= 0;
            end
          else Result:= StrToFloat(m_c_value_list[l_index_of])
      end; // f_amount

    procedure c_calcul_ir.display_parameters;
      var l_ir_parameter_index, l_tranche_index: Integer;
          l_key, l_seuil: String;
          l_key_value_index: Integer;
      begin
        display_line;
        display(m_title);

        l_tranche_index:= 1;
        l_key_value_index:= 0;

        with m_c_key_list do
          for l_ir_parameter_index:= 0 to Count- 1 do
          begin
            l_key:= Strings[l_ir_parameter_index];
            if l_key= ''
              then begin
                  display_line;
                end
              else
                if f_is_tranche(l_key)
                  then begin
                      if m_seuil_array[l_tranche_index]= k_max
                        then l_seuil:= f_spaces(10)
                        else l_seuil:= Format('%10.2f', [m_seuil_array[l_tranche_index] ]);
                      display(Format('%-10s   %8s  %5.2f',
                          [l_key, l_seuil, m_taux_array[l_tranche_index] ]));
                      Inc(l_tranche_index);
                    end
                  else begin
                       display(Format('%-27s   %10.2f %5.2f',
                            [l_key, f_amount(l_key), f_rate(l_key) ]))  ;
                    end;
          end; // fpr
      end; // display_parameters

    procedure c_calcul_ir.compute_ir;
      var l_micro_foncier_net: Double;
          // l_total_csg_crds: Double;

      procedure _calcul_csg_crds_loyers;
        var l_base_flat_tax: Double;

        procedure __display_and_add_csg(p_title: String; p_rate, p_base: Double);
          var l_rate, l_amount: Double;
          begin
            l_amount:= p_base* p_rate/ 100;
            display(Format('%-35s %10s  %5.2f   %10s',
                [p_title, f_10_2(p_base), p_rate, f_10_2(l_amount)]));
            m_total_csg_crds:= m_total_csg_crds+ l_amount;
          end; // __display_and_add_csg

        var l_flat_tax_csg: Double;
            l_full_flat_tax: Double;

        begin // _calcul_csg_crds_loyers
          m_total_csg_crds:= 0;

          display(f_display('micro_foncier_net', l_micro_foncier_net));

          l_flat_tax_csg:= f_rate('flat_tax_CSG')+ f_rate('flat_tax_CRDS')+ f_rate('flat_tax_solidarite');

          display_line;
          __display_and_add_csg('flat_tax_csg', l_flat_tax_csg, l_micro_foncier_net);

          l_base_flat_tax:= 0;
          display_line;
          display(f_display('interets_associes', m_c_calcul_is_ref.m_interets_compte_associe));
          display(f_display('dividendes', m_c_calcul_is_ref.m_dividendes));
          l_base_flat_tax:= m_c_calcul_is_ref.m_interets_compte_associe+  m_c_calcul_is_ref.m_dividendes;

          display_line;
          display(f_display(' = base_csg_crds', l_base_flat_tax));
          display_line;

          l_full_flat_tax:= l_flat_tax_csg+ f_rate('flat_tax_taux_forfaitaire');
          __display_and_add_csg('flat_tax', l_full_flat_tax, l_base_flat_tax);

          display_line;
          display(f_display_2('total_csg_flat_tax', 28, m_total_csg_crds));
(*
          display(Format('total_csg_crds  %s %10.2f  %5.2f %%', [f_spaces(19), l_total_csg_crds, l_total_csg_crds/ l_base_flat_tax]));
*)
        end; // _calcul_csg_crds_loyers

      var l_total_revenus: Double;

      procedure _display_and_add_revenu(p_title: String; p_amount: Double);
        begin
          display(f_display(p_title, p_amount));

          l_total_revenus:= l_total_revenus+ p_amount;
        end; // _display_and_add_revenu

      var l_total_impots: Double;

      procedure _display_and_add_impots(p_title: String; p_amount: Double);
        var l_display: String;
        begin
          l_display:= f_display(p_title, p_amount);

          if m_ir_precedent<> 0
            then
              if pos('IR', p_title)> 0
                then l_display:= l_display+ Format('%s [%10.2f]', [f_spaces(2), p_amount- m_ir_precedent])
                else
                  if Pos('CSG', p_title)> 0
                    then l_display:= l_display+ Format('%s [%10.2f]', [f_spaces(2), p_amount- m_total_csg_crds_precedent]);

          display(l_display);

          l_total_impots:= l_total_impots+ p_amount;
        end; // _display_and_add_impots

      var l_revenu_net_imposable: Double;
          l_revenu_brut_global: Double;
          l_revenu_imposable: Double;

          l_revenu_imposable_par_part: double;
          l_impot_par_tranches: Double;

          l_impot_taux_forfaitaire: Double;

          l_tranche_index: Integer;
          l_base, l_amount, l_total: Double;
          l_seuil: String;

          l_c_save_display_strings: tStrings;

      begin // compute_ir
        display_line;
        display(m_title);
        display_line;

        display(f_display('salaires', m_revenu_imposable));
        if m_csg_deductible_ir> 0
          then display(f_display(' + csg_deductible', m_csg_deductible_ir));
        display(f_display(' + retraites', m_retraites));
        m_revenu_imposable:= m_revenu_imposable+ m_csg_deductible_ir+ m_retraites;

        display_line;

        l_revenu_net_imposable:= 0.90* m_revenu_imposable;
        l_micro_foncier_net:= m_c_calcul_is_ref.m_micro_foncier_brut* 0.70;
        l_revenu_brut_global:= l_revenu_net_imposable+ l_micro_foncier_net;
        l_revenu_imposable:= l_revenu_brut_global- m_csg_deductible_ir;

        display(f_display('=  revenu_imposable', m_revenu_imposable));
        display(f_display(' - frais_10_%% ', 0.10* m_revenu_imposable));
        display(f_display('=  revenu_net_imposable', l_revenu_net_imposable));
        display(f_display(' + micro_foncier_net', l_micro_foncier_net)
             + Format(' (0.70 * foncier %10f.3)', [m_c_calcul_is_ref.m_micro_foncier_brut]));
        display(f_display('= revenu_brut_global', l_revenu_brut_global));
        display(f_display('csg_deductible', m_csg_deductible_ir));
        display(f_display('revenu_imposable', l_revenu_imposable));
        if m_delta_net> 0
          then display(f_display('delta_net_??? ', m_delta_net));

        display_line;

        l_total:= 0;
        l_revenu_imposable:= l_revenu_imposable+ m_delta_net;

        l_c_save_display_strings:= g_c_display.m_c_strings;
        initialize_display(m_c_detail_result_list);
        // clear_display;

        display(f_display('revenu_imposable', l_revenu_imposable));
        l_revenu_imposable_par_part:= l_revenu_imposable/ m_nombre_parts;
        display(f_display('revenu_imposable_par_part', l_revenu_imposable_par_part));
        display_line;
        // l_revenu_imposable_par_part:= 31000* 0.9;

        l_impot_par_tranches:= 0;

        for l_tranche_index:= 1 to m_tranche_count- 1 do
          if l_revenu_imposable_par_part> m_seuil_array[l_tranche_index- 1]
            then begin
                l_base:= Min(l_revenu_imposable_par_part, m_seuil_array[l_tranche_index]);
                l_base:= l_base- m_seuil_array[l_tranche_index- 1];
                l_amount:= l_base* m_taux_array[l_tranche_index]/ 100;
                if m_seuil_array[l_tranche_index]= k_max
                  then l_seuil:= f_spaces(10)
                  else l_seuil:= Format('%10s', [f_10_2(m_seuil_array[l_tranche_index]) ]);
                display(Format('%d %10s  %10s  %6.2f    %10s',
                    [l_tranche_index, l_seuil, f_10_2(l_base), m_taux_array[l_tranche_index], f_10_2(l_amount)]));

                l_total:= l_total+ l_amount;
              end;

        display_line;
        display(Format('%-35s %10s', ['total', f_10_2(l_total)]));

        m_IR:= l_total* m_nombre_parts;
        display_line;
        display(Format('total_pour_%1.1f_parts %s %10s  %3.2f %%',
            [m_nombre_parts, f_spaces(14), f_10_2(m_IR), m_IR/ l_revenu_imposable] ));

        initialize_display(l_c_save_display_strings);

        display(f_display('impot_a_payer', m_IR));


(*

        l_impot_taux_forfaitaire:= m_revenu_au_taux_forfaitaire* f_rate('taux_forfaitaire')/ 100;
        display_line;

        display(Format('revenu_taux_forfaitaire     %10.2f', [m_revenu_au_taux_forfaitaire]));
        display(Format('impottaux_forfaitaire       %10.2f', [l_impot_taux_forfaitaire]));
*)
        l_total:= l_impot_par_tranches+ l_impot_taux_forfaitaire;


        initialize_display(m_c_detail_result_list);
        display_line;
        display_line;

        _calcul_csg_crds_loyers;

        initialize_display(l_c_save_display_strings);
(*
        display_line;

        display('REVENUS');
        l_total_revenus:= 0;
        _display_and_add_revenu('salaire_+_retraites', m_revenu_imposable);
        _display_and_add_revenu(' + micro_foncier_brut', m_c_calcul_is_ref.m_micro_foncier_brut);
        _display_and_add_revenu(' + intrts_compte_assoc', m_c_calcul_is_ref.m_interets_compte_associe);
        _display_and_add_revenu(' + dividendes', m_c_calcul_is_ref.m_dividendes);

        display_line;
        display(f_display('= total_revenus',l_total_revenus));

        display_line;
        display('PRELEVEMENTS');
        l_total_impots:= 0;
        _display_and_add_impots('IR', m_IR);
        _display_and_add_impots(' + CSG', m_total_csg_crds);

        display_line;
        display(f_display('= total_prlvements', l_total_impots));

        display_line;
        display(f_display('RESTE_NET',l_total_revenus- l_total_impots)
            + Format(' %3.2f', [l_total_impots/ (l_total_revenus- l_total_impots)]));

        display_line;
*)
      end; // compute_ir

    Destructor c_calcul_ir.Destroy;
      begin
        m_c_key_list.Free;
        m_c_value_list.Free;

        Inherited;
      end; // Destroy

    begin // u_c_calcul_ir
    end. // u_c_calcul_ir


